import os
import mmap
import hashlib

from ssh_ipykernel.utils import decode_utf8


class Status:
    """Store status of kernel start in mmap'd file for external tools

    Arguments:
        connection_info {dict} -- A ipykernel connection info

    Keyword Arguments:
        status_folder {str} -- Folder where to save the status (default: {"~/.ssh_ipykernel"})
    """

    UNKNOWN = 0
    DOWN = 1
    UNREACHABLE = 2
    KERNEL_KILLED = 3
    STARTING = 4
    RUNNING = 5
    CONNECT_FAILED = 6
    RUNNING_EXT = 7

    MESSAGES = {
        UNKNOWN: "Unknown",
        DOWN: "Cluster down",
        UNREACHABLE: "Cluster unreachable",
        KERNEL_KILLED: "Kernel killed",
        STARTING: "Starting",
        RUNNING: "Running",
        CONNECT_FAILED: "Connect failed",
    }

    ENDIAN = "little"

    def __init__(self, connection_info, logger, status_folder="~/.ssh_ipykernel"):
        self._logger = logger

        self.status_folder = os.path.expanduser(status_folder)
        filename = "%s.status" % self.create_hash(connection_info)
        self.status_file = os.path.join(self.status_folder, filename)
        self.status_available = True
        self.status = self.create_or_get()

    def create_hash(self, connection_info):
        conn_str = "%d-%d-%d-%d-%d" % (
            connection_info["shell_port"],
            connection_info["iopub_port"],
            connection_info["stdin_port"],
            connection_info["control_port"],
            connection_info["hb_port"],
        )
        h = hashlib.sha256()
        h.update(conn_str.encode())
        return h.hexdigest()

    def create_or_get(self):
        """Create the status file

        Returns:
            [mmap] -- Memory mapped status file
        """
        if not os.path.exists(self.status_folder):
            try:
                os.mkdir(self.status_folder)
            except Exception as ex:
                self._logger.error("Cannot create %s" % self.status_folder)
                self._logger.error(str(ex))
                self.status_available = False

        if self.status_available and not os.path.exists(self.status_file):
            self._logger.debug("Creating new status file %s" % self.status_file)
            try:
                with open(self.status_file, "wb") as fd:
                    fd.write((0).to_bytes(12, Status.ENDIAN))
            except Exception as ex:
                self._logger.error("Cannot initialize %s" % self.status_folder)
                self._logger.error(str(ex))
                self.status_available = False

        if self.status_available:
            self._logger.debug("Attaching to status file %s" % self.status_file)
            fd = open(self.status_file, "r+b")
            return mmap.mmap(fd.fileno(), 0)
        else:
            return None

    def _to_bytes(self, value, length):
        return value.to_bytes(length, Status.ENDIAN, signed=False)

    def _from_bytes(self, value):
        return int.from_bytes(value, Status.ENDIAN, signed=False)

    def _set_status(self, status, pid, sudo):
        """Set status if status file exists

        Arguments:
            status {int} -- Status.<value>
        """
        if self.status_available:
            if pid == None:
                pid = self.get_pid()
            if sudo == None:
                sudo = self.is_sudo()

            new_status = self._to_bytes(status, 2) + self._to_bytes(pid, 8)
            new_status += self._to_bytes(1 if sudo else 0, 2)
            self.status[:12] = new_status
            self.status.flush()
            self._logger.debug(
                "Status for remote pid {pid}: {status}".format(
                    status=Status.MESSAGES[status], pid=pid
                )
            )

    def _get_status(self):
        """Get status if status file exists

        Returns:
            int -- Status.<value>
        """
        if self.status_available:
            return self._from_bytes(self.status[:2])
        else:
            return Status.UNKNOWN

    def get_pid(self):
        """Get remote pid of a running ssh_ipykernel

        Returns:
            int -- pid
        """
        if self.status_available:
            return self._from_bytes(self.status[2:10])
        else:
            return -1

    def is_sudo(self):
        """Get sudo state of a running ssh_ipykernel

        Returns:
            bool -- True if sudo is used, else False
        """
        if self.status_available:
            return self._from_bytes(self.status[10:12]) == 1
        else:
            return False

    def set_unreachable(self, pid=None, sudo=None):
        """Set current status to Status.UNREACHABLE
        """
        self._set_status(Status.UNREACHABLE, pid, sudo)

    def set_kernel_killed(self, pid=None, sudo=None):
        """Set current status to Status.KERNEL_KILLED
        """
        self._set_status(Status.KERNEL_KILLED, pid, sudo)

    def set_starting(self, pid=None, sudo=None):
        """Set current status to Status.STARTING
        """
        self._set_status(Status.STARTING, pid, sudo)

    def set_running(self, pid, sudo):
        """Set current status to Status.RUNNING
        """
        self._set_status(Status.RUNNING, pid, sudo)

    def set_running_ext(self, pid, sudo):
        """Set current status to Status.RUNNING_EXT
        """
        self._set_status(Status.RUNNING, pid, sudo)
    
    def set_down(self, pid=None, sudo=None):
        """Set current status to Status.DOWN
        """
        self._set_status(Status.DOWN, pid, sudo)

    def set_connect_failed(self, pid=None, sudo=None):
        """Set current status to Status.CONNECT_FAILED
        """
        self._set_status(Status.CONNECT_FAILED, pid, sudo)

    def is_unknown(self):
        """Check for Status.UNKNOWN

        Returns:
            bool -- True if current status is Status.UNKNOWN
        """
        return self._get_status() == Status.UNKNOWN

    def is_unreachable(self):
        """Check for Status.UNREACHABLE

        Returns:
            bool -- True if current status is Status.UNREACHABLE
        """
        return self._get_status() == Status.UNREACHABLE

    def is_kernel_killed(self):
        """Check for Status.KERNEL_KILLED

        Returns:
            bool -- True if current status is Status.KERNEL_KILLED
        """
        return self._get_status() == Status.KERNEL_KILLED

    def is_starting(self):
        """Check for Status.STARTING

        Returns:
            bool -- True if current status is Status.STARTING
        """
        return self._get_status() == Status.STARTING

    def is_running(self):
        """Check for Status.RUNNING

        Returns:
            bool -- True if current status is Status.RUNNING or Status.RUNNING_EXT
        """
        return self._get_status() in [Status.RUNNING, Status.RUNNING_EXT]

    def is_down(self):
        """Check for Status.DOWN

        Returns:
            bool -- True if current status is Status.DOWN
        """
        return self._get_status() == Status.DOWN

    def is_connect_failed(self):
        """Check for Status.CONNECT_FAILED

        Returns:
            bool -- True if current status is Status.CONNECT_FAILED
        """
        return self._get_status() == Status.CONNECT_FAILED

    def close(self):
        """Close status file if exists
        """
        try:
            if self.status_available:
                os.remove(self.status_file)
            # else:
            #     self._logger.info("no need to delete status file")
        except Exception as ex:
            self._logger.error(str(ex))

    def get_status_message(self):
        """Get human readable versionof status
        """
        return Status.MESSAGES[self._get_status()]



import argparse
import getpass
import json
import os
import re
import sys
import tempfile
from jupyter_client import kernelspec as ks

PREFIX = "ssh_"


def add_kernel(
    host,
    display_name,
    local_python_path,
    remote_python_path,
    env=None,
    sudo=False,
    system=False,
    timeout=5,
    module="ssh_ipykernel",
    opt_args=None,
):
    """Add a new kernel specification for an SSH Kernel

    Arguments:
        host {str} -- host where the remote ipykernel should be started
        display_name {str} -- Display name for the new kernel
        local_python_path {[type]} -- Local python path to be used (without bin/python)
        remote_python_path {[type]} -- Remote python path to be used (without bin/python)

    Keyword Arguments:
        env {str} -- Environment variables passd to the ipykernel "VAR1=VAL1 VAR2=VAL2" (default: {""})
        sudo {bool} -- Start ipykernel as root if necessary (default: {False})
        system {bool} -- Create kernelspec as user (False) or system (True) (default: {False})
        timeout {int} -- SSH connection timeout (default: {5})

    Returns:
        [type] -- [description]
    """

    def simplify(name):
        return re.sub(r"[^a-zA-Z0-9\-\.\_]", "", name)

    if system:
        username = False
    else:
        username = getpass.getuser()

    if opt_args is None:
        opt_args = []

    kernel_json = {
        "argv": [
            local_python_path,
            "-m",
            module,
            "--host",
            host,
            "--python",
            remote_python_path,
            "--timeout",
            str(timeout),
        ]
        + opt_args
        + ["-f", "{connection_file}",],
        "display_name": display_name,
        "language": "python",
    }
    if env is not None:
        kernel_json["argv"].insert(-2, "--env")
        kernel_json["argv"].insert(-2, env)

    if sudo:
        kernel_json["argv"].insert(-2, "-s")

    kernel_name = "{prefix}_{display_name}".format(
        prefix=PREFIX, host=host, display_name=simplify(display_name)
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        os.chmod(temp_dir, 0o755)

        with open(os.path.join(temp_dir, "kernel.json"), "w") as fd:
            json.dump(kernel_json, fd, sort_keys=True, indent=2)

        ks.install_kernel_spec(temp_dir, kernel_name, user=username, replace=True)

    return kernel_name


if __name__ == "__main__":
    parser = argparse.ArgumentParser(add_help=False)
    optional = parser.add_argument_group("optional arguments")
    optional.add_argument(
        "--help",
        "-h",
        action="help",
        default=argparse.SUPPRESS,
        help="show this help message and exit",
    )
    optional.add_argument(
        "--display-name", "-d", type=str, help="kernel display name (default is host name)"
    )
    optional.add_argument(
        "--sudo",
        "-s",
        action="store_true",
        help="sudo required to start kernel on the remote machine",
    )
    optional.add_argument(
        "--timeout", "-t", type=int, help="timeout for remote commands", default=5
    )
    optional.add_argument(
        "--env",
        "-e",
        nargs="*",
        help="environment variables for the remote kernel in the form: VAR1=value1 VAR2=value2",
    )

    required = parser.add_argument_group("required arguments")
    required.add_argument("--host", "-H", required=True, help="remote host")
    required.add_argument("--python", "-p", required=True, help="remote python_path")
    args = parser.parse_args()

    if args.env:
        env = " ".join(args.env)

    add_kernel(
        host=args.host,
        display_name=args.host if args.display_name is None else args.display_name,
        local_python_path=sys.executable,
        remote_python_path=args.python,
        sudo=args.sudo,
        env=env,
        timeout=args.timeout,
    )



import json
import os
from pathlib import Path, PurePosixPath
import platform
import re
import signal
import subprocess
import sys
import uuid

from jupyter_client import BlockingKernelClient
from tornado.log import LogFormatter

from ssh_ipykernel.utils import setup_logging

if platform.system() == "Windows":
    # os.environ["WEXPECT_SPAWN_CLASS"] = "SpawnPipe"
    import wexpect as expect  # pylint: disable=import-error

    # from wexpect.wexpect_util import SIGNAL_CHARS  # pylint: disable=import-error

    is_windows = True
    SSH = "ssh.exe"
    ENCODING = {"codepage": 65001}
    # SIGINT = SIGNAL_CHARS[signal.SIGINT]
else:
    import pexpect as expect  # pylint: disable=import-error

    is_windows = False
    SSH = "ssh"
    ENCODING = {"encoding": "utf-8"}
    # SIGINT = signal.SIGINT

from .status import Status


# No tabs, no multiline, quote { and } !
KERNEL_SCRIPT = """
import json
import os
fname = os.path.expanduser("{fname}")
from jupyter_client import write_connection_file
write_connection_file(fname=fname, ip="{ip}", key=b"{key}", transport="{transport}", signature_scheme="{signature_scheme}", kernel_name="{kernel_name}")
fd = open(fname, "r")
ci = json.loads(fd.read())
fd.close()
ports = json.dumps({{k:v for k,v in ci.items() if "_port" in k}})
print(ports)
"""


class SshKernelException(Exception):
    pass


class SshKernel:
    """Remote ipykernel via SSH

    Raises:
        SshKernelException: "Could not execute remote command, connection died"
        SshKernelException: "Connection failed"
        SshKernelException: "Could not create kernel_info file"

        Arguments:
            host {str} -- host where the remote ipykernel should be started
            connection_info {dict} -- Local ipykernel connection info as provided by Juypter lab
            python_path {str} -- Remote python path to be used to start ipykernel

        Keyword Arguments:
            sudo {bool} -- Start ipykernel as root if necessary (default: {False})
            timeout {int} -- SSH connection timeout (default: {5})
            env {str} -- Environment variables passd to the ipykernel "VAR1=VAL1 VAR2=VAL2" (default: {""})
            ssh_config {str} -- Path to the local SSH config file (default: {Path.home() / ".ssh" / "config"})
    """

    def __init__(
        self,
        host,
        connection_info,
        python_path,
        sudo=False,
        timeout=5,
        env="",
        ssh_config=None,
        quiet=True,
        verbose=False,
        msg_interval=30,
        logger=None,
    ):
        self.host = host
        self.connection_info = connection_info
        self.python_full_path = PurePosixPath(python_path) / "bin/python"
        self.sudo = sudo
        self.timeout = timeout
        self.env = env
        self.ssh_config = (
            Path.home() / ".ssh" / "config" if ssh_config is None else ssh_config
        )  # OS specific path

        self.quiet = quiet
        self.verbose = verbose

        self._connection = None

        self.remote_ports = {}
        self.uuid = str(uuid.uuid4())
        self.fname = "/tmp/.ssh_ipykernel_%s.json" % self.uuid  # POSIX path

        if logger is None:
            self._logger = setup_logging("SshKernel")
        else:
            self._logger = logger

        self._logger.debug("Remote kernel info file: {0}".format(self.fname))
        self._logger.debug("Local connection info: {0}".format(connection_info))

        self.kernel_pid = 0
        self.status = Status(connection_info, self._logger)
        self.msg_interval = int(msg_interval / timeout)
        self.msg_counter = 0

    def _execute(self, cmd):
        try:
            result = subprocess.check_output(cmd)
            return 0, result
        except subprocess.CalledProcessError as e:
            return e.returncode, e.args

    def _ssh(self, cmd):
        return self._execute([SSH, self.host, cmd])

    def close(self):
        """Close pcssh connection
        """
        if self._connection is not None:  # and self._connection.isalive():
            if self._connection.isalive():
                self._connection.logout()
                self._logger.debug("Ssh connection closed")
            if self.kc.is_alive():
                self.kc.stop_channels()
                self._logger.debug("Kernel client channels stopped")

    def create_remote_connection_info(self):
        """Create a remote ipykernel connection info file
        Uses KERNEL_SCRIPT to execute jupyter_client.write_connection_file remotely to request remote ports.
        The remote ports will be returned as json and stored to built the SSH tunnels later.
        The pxssh connection will be closed at the end.

        Raises:
            SshKernelException: "Could not create kernel_info file"
        """
        self._logger.info("Creating remote connection info")
        script = KERNEL_SCRIPT.format(fname=self.fname, **self.connection_info)

        cmd = "{python} -c '{command}'".format(
            python=self.python_full_path, command="; ".join(script.strip().split("\n"))
        )

        result = self._ssh(cmd)
        self._logger.debug(result)
        if result[0] == 0:
            self.remote_ports = json.loads(result[1].decode("utf-8"))
            self._logger.debug(
                "Local ports  = %s"
                % {k: v for k, v in self.connection_info.items() if "_port" in k}
            )
            self._logger.debug("Remote ports = %s" % self.remote_ports)
        else:
            self.status.set_unreachable(self.kernel_pid, self.sudo)
            raise SshKernelException("Could not create kernel_info file")

    def kernel_client(self):
        self.kc = BlockingKernelClient()
        self.kc.load_connection_info(self.connection_info)
        self.kc.start_channels()

    def kernel_init(self):
        done = False
        if self.check_alive(show_pid=False):
            i = 0
            while not done:
                try:
                    i += 1
                    self._logger.debug("Retrieving kernel pid, attempt %d" % i)
                    result = self.kc.execute_interactive(
                        "import os",
                        user_expressions={"pid": "os.getpid()"},
                        store_history=False,
                        silent=True,
                        timeout=2,
                    )
                    self._logger.debug("result = %s" % str(result["content"]))
                    self.kernel_pid = int(
                        result["content"]["user_expressions"]["pid"]["data"]["text/plain"]
                    )
                    self._logger.debug("Remote kernel pid %d" % self.kernel_pid)
                    done = True
                except Exception as ex:
                    msg = str(ex)
                    if msg == "Timeout waiting for output":
                        self._logger.warning("Warning: {}".format(msg))
                        if i > 5:
                            self._logger.error("Max attempts (5) reached, stopping")
                            raise SshKernelException("Could not initialize kernel")
                            break
                    else:
                        self._logger.error("Warning: {}".format(str(ex)))
        return done

    def kernel_customize(self):
        pass

    def check_alive(self, show_pid=True):
        alive = self._connection.isalive() and self.kc.is_alive()
        if show_pid:
            msg = "Remote kernel ({}, pid = {}) is {}alive".format(
                self.host, self.kernel_pid, "" if alive else "not "
            )
        else:
            msg = "Remote kernel is {}alive".format("" if alive else "not ")

        if not alive or self.msg_counter % self.msg_interval == 0:
            self.msg_counter = 0
            self._logger.info(msg)

        self.msg_counter += 1
        return alive

    def interrupt_kernel(self):
        if self._connection.isalive():
            if is_windows:
                self._logger.warning('On Windows use "Interrupt remote kernel" button')
            else:
                self._logger.warning("Sending interrupt to remote kernel")
                self._connection.sendintr()  # send SIGINT

    def start_kernel_and_tunnels(self):
        """Start Kernels and SSH tunnels
        A new pxssh connection will be created that will
        - set up the necessary ssh tunnels between remote kernel ports and local kernel ports
        - start the ipykernel on the remote host
        """
        self._logger.info("Setting up ssh tunnels")

        ssh_tunnels = []
        for port_name in self.remote_ports.keys():
            ssh_tunnels += [
                "-L",
                "{local_port}:127.0.0.1:{remote_port}".format(
                    local_port=self.connection_info[port_name],
                    remote_port=self.remote_ports[port_name],
                ),
            ]

        self._logger.info("Starting remote kernel")

        # Build remote command
        sudo = "sudo " if self.sudo else ""

        if self.env is not None:
            env = " ".join(self.env)
        cmd = "{sudo} {env} {python} -m ipykernel_launcher -f {fname}".format(
            sudo=sudo, env=env, python=self.python_full_path, fname=self.fname
        )

        # Build ssh command with all flags and tunnels
        if self.quiet:
            args = ["-q"]
        elif self.verbose:
            args = ["-v"]
        else:
            args = []
        args += ["-t", "-F", str(self.ssh_config)] + ssh_tunnels + [self.host, cmd]

        self._logger.debug("%s %s" % (SSH, " ".join(args)))

        try:
            # Start the child process
            self._connection = expect.spawn(SSH, args=args, timeout=self.timeout, **ENCODING)
            # subprocess.check_output([SSH] + args)
            #
            # get blocking kernel client
            self.kernel_client()
            # initialize it
            if self.kernel_init():
                self.status.set_running(self.kernel_pid, self.sudo)
                # run custom code if part of sub class
                self.kernel_customize()
            else:
                self.status.set_connect_failed(sudo=self.sudo)
        except Exception as e:
            tb = sys.exc_info()[2]
            self._logger.error(str(e.with_traceback(tb)))
            self._logger.error("Cannot contiune, exiting")
            sys.exit(1)

        prompt = re.compile(r"\n")

        while True:
            try:
                # Wait for prompt
                self._connection.expect(prompt)
                # print the outputs
                self._logger.info(self._connection.before.strip("\r\n"))

            except KeyboardInterrupt:
                self.interrupt_kernel()
                self.check_alive()

            except expect.TIMEOUT:
                self.check_alive()

            except expect.EOF:
                # The program has exited
                self._logger.info("The program has exited.")
                self.status.set_down(self.kernel_pid, self.sudo)
                break

        self.close()
        self.status.close()



# -*- coding: utf-8 -*-

"""Top-level package for SSH Kernel."""

from notebook.utils import url_path_join

__author__ = """Bernhard Walter"""
__email__ = "b_walter@arcor.de"
from ._version import __version__, __version_info__

from .ssh_ipykernel_interrupt import SshInterruptHandler


def load_jupyter_server_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    web_app = nb_server_app.web_app
    SshInterruptHandler.nbapp = nb_server_app
    host_pattern = ".*$"
    route_pattern = url_path_join(web_app.settings["base_url"], "/interrupt")
    web_app.add_handlers(host_pattern, [(route_pattern, SshInterruptHandler)])



import argparse
import json
import sys
from .kernel import SshKernel


def main(host, connection_info, python_path, sudo, timeout, env):
    """Main function to be called as module to create SshKernel

    Arguments:
        host {str} -- host where the remote ipykernel should be started
        connection_info {dict} -- Local ipykernel connection info as provided by Juypter lab
        python_path {str} -- Remote python path to be used to start ipykernel
        sudo {bool} -- Start ipykernel as root if necessary (default: {False})
        timeout {int} -- SSH connection timeout (default: {5})
        env {str} -- Environment variables passd to the ipykernel "VAR1=VAL1 VAR2=VAL2" (default: {""})
    """
    kernel = SshKernel(host, connection_info, python_path, sudo, timeout, env)
    try:
        kernel.create_remote_connection_info()
        kernel.start_kernel_and_tunnels()
    except:
        kernel._logger.error("Kernel could not be started")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(add_help=False)
    #    parser._action_groups.pop()
    optional = parser.add_argument_group("optional arguments")
    optional.add_argument(
        "--help",
        "-h",
        action="help",
        default=argparse.SUPPRESS,
        help="show this help message and exit",
    )
    optional.add_argument(
        "--timeout", "-t", type=int, help="timeout for remote commands", default=5
    )
    optional.add_argument(
        "--env",
        "-e",
        nargs="*",
        help="environment variables for the remote kernel in the form: VAR1=value1 VAR2=value2",
    )
    optional.add_argument(
        "-s", action="store_true", help="sudo required to start kernel on the remote machine"
    )

    required = parser.add_argument_group("required arguments")
    required.add_argument("--file", "-f", required=True, help="jupyter kernel connection file")
    required.add_argument("--host", "-H", required=True, help="remote host")
    required.add_argument("--python", "-p", required=True, help="remote python_path")
    args = parser.parse_args()

    try:
        with open(args.file, "r") as fd:
            connection_info = json.loads(fd.read())
    except Exception as ex:
        print(ex)
        sys.exit(1)

    sys.exit(main(args.host, connection_info, args.python, args.s, args.timeout, args.env))



import logging
import os
import platform
import subprocess
from tornado.log import LogFormatter


if platform.system() == "Windows":
    is_windows = True
    SSH = "ssh.exe"
else:
    is_windows = False
    SSH = "ssh"


def setup_logging(name):
    """Setup Logging
    """
    _log_fmt = (
        "%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d " "%(name)s]%(end_color)s %(message)s"
    )
    _log_datefmt = "%H:%M:%S"

    logger = logging.getLogger(name)
    debug_level = os.environ.get("DEBUG", "INFO")
    logger.setLevel(debug_level)
    logger.propagate = False
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    console.setFormatter(LogFormatter(fmt=_log_fmt, datefmt=_log_datefmt))
    logger.addHandler(console)
    return logger


logger = setup_logging("ssh_ipykernel:utils")


def execute(cmd):
    try:
        logger.debug("interrupt cmd = %s" % cmd)
        result = subprocess.check_output(cmd)
        result = {"code": 0, "data": result.decode("utf-8")}
    except subprocess.CalledProcessError as e:
        result = {"code": e.returncode, "data": e.args}

    logger.debug("result=%s", str(result))
    return result


def ssh(host, cmd):
    return execute([SSH, host, cmd])


def decode_utf8(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return s.decode("utf-8", "replace")
    else:
        raise ValueError("s is neither str nor bytes")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import platform
from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

setup(
    author="Bernhard Walter",
    author_email="b_walter@arcor.de",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A remote jupyter ipykernel via ssh",
    install_requires=[
        "tornado>=6.1.0,<=6.2.0",
        "jupyter_client>=6.1.12,<6.2.0",
        "jupyterlab>=3.0.0,<3.1.0",
        "wexpect==3.3.2;platform_system=='Windows'",
        "pexpect==4.8.0;platform_system!='Windows'",
        "ssh_ipykernel_interrupt==1.1.0",
    ],
    extras_require={
        "dev": {"twine", "bumpversion", "black", "pylint", "wheel"},
    },
    data_files=[
        (
            "etc/jupyter/jupyter_server_config.d",
            ["ssh_ipykernel/ssh_ipykernel_interrupt/jupyter_server/ssh-ipykernel-interrupt.json"],
        ),
        # (
        #     "etc/jupyter/jupyter_notebook_config.d",
        #     ["ssh_ipykernel/ssh_ipykernel_interrupt/jupyter_notebook/ssh-ipykernel-interrupt.json"],
        # ),
    ],
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    # packages=find_packages(),
    include_package_data=True,
    keywords="ssh_ipykernel",
    name="ssh_ipykernel",
    packages=find_packages(exclude=["ssh_ipykernel_interrupt"]),
    python_requires=">=3.5",
    url="https://github.com/bernhard-42/ssh_ipykernel",
    version="1.2.2",
    zip_safe=False,
)



import json
import os
import signal

from notebook.base.handlers import IPythonHandler
from tornado import web

from ssh_ipykernel.utils import ssh, setup_logging

from ssh_ipykernel.status import Status

logger = setup_logging("ssh_ipykernel:interrupt")


class SshInterruptHandler(IPythonHandler):
    """Kernel handler to interrupt remote ssh ipykernel"""

    nbapp = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_kernel(self, kernel_id):
        """Get jupyter kernel for given kernel Id
        
        Args:
            kernel_id (str): Internal jupyter kernel ID
        
        Returns:
            KernelManager: KernelManager object
        """
        km = SshInterruptHandler.nbapp.kernel_manager

        kernel_info = None
        for kernel_info in km.list_kernels():
            if kernel_info["id"] == kernel_id:
                break

        if kernel_info is not None:
            return km.get_kernel(kernel_info["id"])
        else:
            return None

    @web.authenticated
    def get(self):
        """GET handler to interrrupt remote ssh ipykernel"""

        kernel_id = self.get_argument("id", None, True)
        logger.debug("kernel id %s" % kernel_id)
        kernel = self.get_kernel(kernel_id)

        status = Status(kernel.get_connection_info(), logger)
        if status.is_running():
            pid = status.get_pid()

            host = ""
            for i, v in enumerate(kernel.kernel_spec.argv):
                if v == "--host":
                    host = kernel.kernel_spec.argv[i + 1]
                    break

            logger.warning("Interrupt remote kernel ({}, pid = {})".format(host, pid))

            cmd = "kill -{sig} {pid}".format(sig=signal.SIGINT.real, pid=pid)
            if status.is_sudo():
                cmd = "sudo " + cmd
            result = ssh(host, cmd)
        else:
            result = {"code": -1, "data": "Remote kernel not running"}

        self.finish(json.dumps(result))


def _jupyter_server_extension_paths():
    """
    Set up the server extension for status handling
    """
    return [{"module": "ssh_ipykernel",}]



from collections import namedtuple
import re

VersionInfo = namedtuple("VersionInfo", ["major", "minor", "patch", "release", "build"])


def get_version(version):
    r = re.compile(
        r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-{0,1}(?P<release>\D*)(?P<build>\d*)"
    )
    major, minor, patch, release, build = r.match(version).groups()
    return VersionInfo(major, minor, patch, release, build)


__version__ = "1.2.2"  # DO NOT EDIT THIS DIRECTLY!  It is managed by bumpversion
__version_info__ = get_version(__version__)
